from __future__ import annotations

import datetime
from pathlib import Path
from typing import Iterable, Iterator, List
import logging.handlers

from octoprint.util import get_dos_filename

from .ftps_client import IoTFTPSClient, IoTFTPSConnection
from .file_info import FileInfo


class RemoteSDCardFileList:

    def __init__(self, settings) -> None:
        self._settings = settings
        self._selected_project_file: FileInfo | None = None
        self._logger = logging.getLogger("octoprint.plugins.bambu_printer.BambuPrinter")
        self._mock_files = []  # Lokales Cache für Mock-Dateien

    def delete_file(self, file_path: Path) -> None:
        try:
            with self.get_ftps_client() as ftp:
                if ftp.delete_file(file_path.as_posix()):
                    self._logger.debug(f"{file_path} deleted")
                else:
                    raise RuntimeError(f"Deleting file {file_path} failed")
        except Exception as e:
            self._logger.exception(e)

    def list_files(
        self,
        folder: str,
        extensions: str | list[str] | None,
        ftp: IoTFTPSConnection,
        existing_files=None,
    ):
        if existing_files is None:
            existing_files = []

        return list(
            self.get_file_info_for_names(
                ftp, ftp.list_files(folder, extensions), existing_files
            )
        )

    def _get_ftp_file_info(
        self,
        ftp: IoTFTPSConnection,
        file_path: Path,
        existing_files: list[str] | None = None,
    ):
        file_size = ftp.get_file_size(file_path.as_posix())
        date = ftp.get_file_date(file_path.as_posix())
        file_name = file_path.name.lower()
        dosname = get_dos_filename(file_name, existing_filenames=existing_files).lower()
        return FileInfo(
            dosname,
            file_path,
            file_size if file_size is not None else 0,
            date,
        )

    def get_file_info_for_names(
        self,
        ftp: IoTFTPSConnection,
        files: Iterable[Path],
        existing_files: list[str] | None = None,
    ) -> Iterator[FileInfo]:
        if existing_files is None:
            existing_files = []

        for entry in files:
            try:
                file_info = self._get_ftp_file_info(ftp, entry, existing_files)
                yield file_info
                existing_files.append(file_info.file_name)
                existing_files.append(file_info.dosname)
            except Exception as e:
                self._logger.exception(e, exc_info=False)

    def get_ftps_client(self):
        """
        Implementieren wir eine Mock-Version des FTPS-Clients, die keinen echten FTP-Zugriff erfordert.
        """
        class MockFTPSClient:
            def __enter__(self):
                return self
                
            def __exit__(self, exc_type, exc_val, exc_tb):
                pass
                
            def get_file_list(self, path=""):
                """Gibt die Mock-Dateiliste zurück"""
                return self._mock_files
                
        mock_client = MockFTPSClient()
        mock_client._mock_files = self._mock_files
        return mock_client
        
    @property
    def is_available(self) -> bool:
        """
        Da wir kein FTP verwenden, ist dieser Service immer verfügbar
        """
        return True

    def get_file_list(self, path: str) -> List[FileInfo]:
        """
        Gibt eine Liste von Dateien im angegebenen Pfad zurück.
        Da wir kein FTP verwenden, geben wir eine leere Liste oder gespeicherte Mock-Dateien zurück.
        """
        self._logger.debug(f"Listing files in path: {path}")
        return self._mock_files

    def add_mock_file(self, file_info: FileInfo):
        """
        Fügt eine Mock-Datei zur Liste hinzu (für Tests oder wenn keine FTP-Verbindung möglich ist)
        """
        self._mock_files.append(file_info)
        self._logger.debug(f"Added mock file: {file_info.file_name}")
    
    def clear_mock_files(self):
        """Löscht alle gespeicherten Mock-Dateien"""
        self._mock_files = []
        self._logger.debug("Mock file list cleared")

    def delete_file(self, path: str) -> bool:
        """
        Simuliert das Löschen einer Datei, entfernt sie aus der Mock-Liste
        """
        self._logger.debug(f"Deleting file: {path}")
        before_count = len(self._mock_files)
        self._mock_files = [f for f in self._mock_files if f.path != path]
        return before_count > len(self._mock_files)
